@@ -21,11 +21,11 @@
 #########################################################################
 
 from asyncworker import AsyncWorker, AsyncItem, PriorityQueue
-from time import time
+from time import time, mktime
 from utils import get_feed_id, get_feed_icon_uri, get_feed_subtitle
 from Queue import Queue
 import constants
-import cPickle
+import cPickle as pickle
 import feedparser
 import gobject
 import gtk
@@ -102,6 +102,14 @@ class ReSiStanceFeedDict(feedparser.FeedParserDict):
         if 'sync' in self:
             del self['sync']
 
+class GoogleReaderSyncData(dict):
+
+    def __init__(self, last_sync, last_update):
+        super(dict, self).__init__()
+
+        self.last_sync = last_sync
+        self.last_update = last_update
+
 class ReSiStanceFeedSummary(dict):
 
     def __init__(self, feed_id, href, title, subtitle, favicon, visits, unread, sync):
@@ -140,11 +148,18 @@ class FeedManager(gobject.GObject):
         self._conn_manager.connect('connection-changed', self._on_connection_changed)
         self._google_reader_auth = None
         self._google_reader_auth_token = None
-        self._google_reader_expires = None
         self._async_worker = None
         self._label_dict = {}
         self._label_dict_loaded = False
 
+        try:
+            db_file = open(constants.RSS_DB_GOOGLE_READER_SYNC_FILE, 'r')
+            self._google_reader_last_sync_dict = pickle.load(db_file)
+            db_file.close()
+        except IOError, e:
+            print e
+            self._google_reader_last_sync_dict = {}
+
     def stop(self):
         # Cancel all pending operations
         if self._async_worker:
@@ -172,6 +187,20 @@ class FeedManager(gobject.GObject):
     def _on_connection_changed(self, conn_manager):
         pass
 
+    def _maybe_create_google_reader_sync_data(self, feed_id, feed_data):
+        try:
+            if 'feed_id' not in self._google_reader_last_sync_dict:
+                last_time = feed_data.feed.updated_parsed if 'updated_parsed' in feed_data.feed \
+                    else feed_data.entries[0].updated_parsed
+                first_time = feed_data.entries[-1].updated_parsed
+
+                self._google_reader_last_sync_dict[feed_id] = \
+                    GoogleReaderSyncData(int(mktime(first_time)), int(mktime(last_time)))
+        except AttributeError:
+            # Init to 1 month ago in case of error
+            one_month_ago = int(time()) - 30*24*3600
+            self._google_reader_last_sync_dict[feed_id] = GoogleReaderSyncData(one_month_ago, one_month_ago)
+
     @requireWorker
     def add_feed(self, url, sync, callback, data=None):
         # Create a new worker to update from network
@@ -212,6 +241,9 @@ class FeedManager(gobject.GObject):
         else:
             pixbuf = self._get_favicon_sync(new_feed_data.href)
 
+        if (sync):
+            self._maybe_create_google_reader_sync_data(new_feed_id, new_feed_data)
+
         self.emit('feed-added', new_feed_id)
 
         return pixbuf, new_feed_data
@@ -232,10 +264,9 @@ class FeedManager(gobject.GObject):
                 print 'ERROR: cannot load feed with ID: ' + string(feed_id)
                 return
 
-        try:
-            updated_feed_data = ReSiStanceFeedDict(feedparser.parse(feed_data.href))
-        except e:
-            return
+        updated_feed_data = ReSiStanceFeedDict(feedparser.parse(feed_data.href))
+        if 'bozo_exception' in updated_feed_data:
+            raise updated_feed_data.bozo_exception
 
         # In case of network failure
         if updated_feed_data == None or updated_feed_data.entries == None or \
@@ -251,6 +282,10 @@ class FeedManager(gobject.GObject):
         if updated_feed_date <= feed_date:
             return
 
+        # Update the sync management data
+        if self._feeds_summary[feed_id].sync:
+            self._google_reader_last_sync_dict[feed_id].last_update = int(mktime(updated_feed_date))
+
         # Use the entry link for those entries with no id attribute
         old_entry_ids = [entry.id if 'id' in entry else entry.link \
                              for entry in feed_data.entries]
@@ -549,6 +584,15 @@ class FeedManager(gobject.GObject):
 
     def save_sync(self):
 
+        # Save Google Reader sync data
+        try:
+            db_file = open(constants.RSS_DB_GOOGLE_READER_SYNC_FILE, 'w')
+            pickle.dump(self._google_reader_last_sync_dict, db_file)
+            db_file.close()
+        except:
+            print 'Could not save google reader sync data'
+            pass
+
         # Save labels, summaries and feed data
         if self._label_dict_loaded:
             try:
@@ -569,7 +613,7 @@ class FeedManager(gobject.GObject):
             # TODO: migrate to "with" statement when available
             try:
                 db_file = open(file_name, 'w')
-                cPickle.dump(feed, db_file)
+                pickle.dump(feed, db_file)
                 db_file.close()
             except:
                 print 'Could not save feed ' + feed.href
@@ -582,19 +626,25 @@ class FeedManager(gobject.GObject):
         self._async_worker.add_task((self.DEFAULT_PRIORITY, load_item))
 
     def _load_in_thread(self, feed_id):
+
+        summary = self._feeds_summary[feed_id]
         try:
             db_file = open(os.path.join(constants.RSS_DB_DIR, str(feed_id)), 'r')
-            feed_data = cPickle.load(db_file)
+            feed_data = pickle.load(db_file)
             db_file.close()
         except IOError, e:
             print e
             # If the data does not exist reload it
-            summary = self._feeds_summary[feed_id]
             feed_data = ReSiStanceFeedDict(feedparser.parse(summary.href))
         except Exception:
             return None
 
         self._feed_data_dict[feed_id] = feed_data
+
+        # Update the sync management data if needed
+        if summary.sync:
+            self._maybe_create_google_reader_sync_data(feed_id, feed_data)
+
         return feed_data
 
     @requireWorker
@@ -690,92 +740,73 @@ class FeedManager(gobject.GObject):
         if not auth_token:
             return False
 
-        # Do not update if we have updated just a while ago
-        if time() < self._google_reader_expires:
+        # Get feeds to sync. Ignore those not synced with Google Reader.
+        all_sync_feed_ids = [get_feed_id(feed_data) for feed_data in self._feed_data_dict.values() \
+                                 if self._feeds_summary[get_feed_id(feed_data)].sync]
+
+        # If last_update <= last_sync it means that feed data has not
+        # been updated since the last time it was synchronized and
+        # thus we can ignore that feed last updated time
+        feed_ids_to_sync = [feed_id for feed_id in all_sync_feed_ids if \
+                                self._google_reader_last_sync_dict[feed_id].last_update > \
+                                self._google_reader_last_sync_dict[feed_id].last_sync]
+        if not feed_ids_to_sync:
             return True
 
-        try:
-            # Get items from feed excluding those with read status. It
-            # indeed really sucks that we can only ask for the unread
-            # items from *ALL* of our subscriptions
-            unread_url = constants.URL_USER + constants.STATE_SUFFIX + 'reading-list' + '?' + 'xt=user/-/' + constants.STATE_SUFFIX + 'read' + '&n=1000'
-            unread_request = urllib2.Request(unread_url, None, auth_headers)
-            unread_response = urllib2.urlopen(unread_request)
-            # Set 10 minutes expiration time. Google returns some
-            # expiration time in unread_response.info()['expires'] but
-            # it's just some seconds.
-            self._google_reader_expires = time() + 10 * 60
-            unread_data = unread_response.read()
-        except:
-            return False
-
-        unread_dict = {}
-        while True:
-            doc = minidom.parseString(unread_data)
-            nodesEntries = doc.getElementsByTagName('entry')
-
-            for node in nodesEntries:
-                entry_link = node.getElementsByTagName('link')[0].attributes['href'].value
-                if not node.getElementsByTagName('source')[0].getAttribute('gr:stream-id').startswith('feed/'):
-                    continue
-
-                feed_source = node.getElementsByTagName('source')[0].getElementsByTagName('link')[0].attributes['href'].value
-                if not feed_source in unread_dict:
-                    unread_dict[feed_source] = [ entry_link ]
-                else:
-                    unread_dict[feed_source].append(entry_link)
-
+        for feed_id in feed_ids_to_sync:
+            now = int(time())
+            unread_list = []
+            summary = self._feeds_summary[feed_id]
+            start_time = self._google_reader_last_sync_dict[feed_id].last_sync
             try:
-                nodes = doc.getElementsByTagName('gr:continuation')
-                if nodes==[]:
-                    break
-                unread_request = urllib2.Request(unread_url+'&c='+nodes[0].firstChild.data, None, auth_headers)
+                # Get items from feed excluding those with read status.
+                unread_url = constants.URL_FEED + summary.href + '?' + \
+                    'xt=user/-/' + constants.STATE_SUFFIX + 'read' + '&n=100' + '&r=o' \
+                    '&ot=' + str(start_time) + '&ck=' + str(now)
+                print unread_url
+
+                unread_request = urllib2.Request(unread_url, None, auth_headers)
                 unread_response = urllib2.urlopen(unread_request)
                 unread_data = unread_response.read()
+
+                while True:
+                    doc = minidom.parseString(unread_data)
+                    nodesEntries = doc.getElementsByTagName('entry')
+
+                    for node in nodesEntries:
+                        entry_link = node.getElementsByTagName('link')[0].attributes['href'].value
+                        if not node.getElementsByTagName('source')[0].getAttribute('gr:stream-id').startswith('feed/'):
+                            continue
+                        unread_list.append(entry_link)
+
+                    try:
+                        nodes = doc.getElementsByTagName('gr:continuation')
+                        if nodes==[]:
+                            break
+                        unread_request = urllib2.Request(unread_url+'&c='+nodes[0].firstChild.data, None, auth_headers)
+                        unread_response = urllib2.urlopen(unread_request)
+                        unread_data = unread_response.read()
+                    except:
+                        break
             except:
-                return False
-
-        # Load all data. FIXME: there should be a better way
-        for summary in self._feeds_summary.values():
-            if summary.feed_id not in self._feed_data_dict:
-                feed_data = self._load_in_thread(summary.feed_id)
-
-        for key in unread_dict.keys():
-            # This could happen if there are unread items in feeds
-            # that are not synchronized to ReSiStance. If that's the
-            # case just ignore them
-            feed_data_list = [feed_data for feed_data in self._feed_data_dict.values() \
-                                  if feed_data.feed.link == key]
-            if not feed_data_list:
                 continue
 
-            feed_data = feed_data_list[0]
+            # Update the last_sync status
+            sync_data = self._google_reader_last_sync_dict[feed_id]
+            if sync_data.last_sync < sync_data.last_update:
+                self._google_reader_last_sync_dict[feed_id].last_sync = now
 
             # Sync read/unread status. Ideally this should prioritize
             # the last action. Meanwhile prioritize the read status
+            feed_data = self.get_feed_data(feed_id)
             for entry in feed_data.entries:
                 # Sometimes entry links include queries, discard them for the comparison
-                if entry.link in unread_dict[key] or entry.link[:entry.link.find('?')] in unread_dict[key]:
+                if entry.link in unread_list or entry.link[:entry.link.find('?')] in unread_list:
                     # If read in ReSiStance update Google Reader (most likely read while offline)
                     if entry.read:
                         self._mark_as_read_synchronize_in_thread(feed_data.feed.link, entry.link, True)
                 else:
                     # If read in Google Reader update ReSiStance (most likely read in Web)
-                    if not entry.read:
-                        entry.read = True
-
-        # Mark all entries as read for feeds synced with Google Reader
-        # with no unread items
-        sync_feeds = []
-        for summary in self._feeds_summary.values():
-            if summary.sync:
-                feed_data = self._feed_data_dict[summary.feed_id]
-                if feed_data.feed.link not in unread_dict:
-                    sync_feeds.append(feed_data)
-
-        for feed_data in sync_feeds:
-            for entry in feed_data.entries:
-                if not entry.read:
                     entry.read = True
 
         return True
@@ -807,7 +838,7 @@ class FeedManager(gobject.GObject):
 
     def _create_summary_from_old_storage(self):
         db_file = open(constants.OLD_RSS_DB_FILE, 'r')
-        feed_data_list = cPickle.load(db_file)
+        feed_data_list = pickle.load(db_file)
         db_file.close()
 
         for feed_data in feed_data_list:
@@ -821,7 +852,7 @@ class FeedManager(gobject.GObject):
     def _load_feeds_summary_in_thread(self):
         try:
             db_file = open(constants.RSS_DB_SUMMARY_FILE, 'r')
-            self._feeds_summary = cPickle.load(db_file)
+            self._feeds_summary = pickle.load(db_file)
             self._feeds_summary_loaded = True
             db_file.close()
         except IOError, e:
@@ -860,7 +891,7 @@ class FeedManager(gobject.GObject):
                 summary_list[0].update_from_feed_data(feed_data)
 
         db_file = open(constants.RSS_DB_SUMMARY_FILE, 'w')
-        cPickle.dump(self._feeds_summary, db_file)
+        pickle.dump(self._feeds_summary, db_file)
         db_file.close()
 
     # LABELS
@@ -906,7 +937,7 @@ class FeedManager(gobject.GObject):
     def _load_labels_in_thread(self):
         try:
             db_file = open(constants.RSS_LABEL_DB_FILE, 'r')
-            self._label_dict = cPickle.load(db_file)
+            self._label_dict = pickle.load(db_file)
             db_file.close()
             loaded = True
         except IOError:
@@ -919,5 +950,5 @@ class FeedManager(gobject.GObject):
 
     def _save_labels(self):
         db_file = open(constants.RSS_LABEL_DB_FILE, 'w')
-        cPickle.dump(self._label_dict, db_file)
+        pickle.dump(self._label_dict, db_file)
         db_file.close()
